﻿
using System.Collections.Generic;
using System.Linq;
using System.Drawing;
using System;
namespace WordTree
{
    /*DISPLAY LIST PAKET*/
    
    /*  Dient zum Darstellen von fluessigen Interpolationen zwischen zwei Baeumen.
     *  Es gibt den Start-Baum und den Ziel-Baum.
     *  Woerter und Beziers jeden Baums sind eigene Knoten (DisplayNode) in der Display List
     *  
     *  Alle Koordinaten und Distanzen sind in NICHT-Zoom Koordinaten gespeichert.
     */

    /**
     * Upper class for display list nodes.
     */
    public class DisplayNode
    {
        public float sourceBeginX; //x,y vom linken oberen Ende des Source-Elemens
        public float sourceBeginY;
        public float targetBeginX; //x,y vom linken oberen Ende des Target-Elements
        public float targetBeginY;
        public float currentBeginX; //x,y vom linken oberen Ende des aktuellen Interpolationsschritts
        public float currentBeginY;
    }

    /**
     * Special display node, which displays a word.
     */
    public class DisplayNodeText : DisplayNode
    {
        public string word; //das wort als string
        public float sourceFontsize; //die Schriftgroesse des Worts im Source-Baum
        public float targetFontsize; //die Schriftgroesse des Worts im Target-Baum
        public float currentFontsize; //die Schriftgroesse des akt. Interpolationsschritts

        /**
         * Constructs this node.
         */
        public DisplayNodeText(string _word, PointF startPosition, float _fontsize)
        {
            word = _word;
            sourceBeginX = startPosition.X;
            sourceBeginY = startPosition.Y;
            sourceFontsize = _fontsize;
        } 
    }

    /**
     * Special display node, which displays a bezier curve (4 control points, but only defined by 2).
     */
    public class DisplayNodeBezier : DisplayNode
    {
        public float sourceEndX; //Endpunkt xy des Beziers im Sourcebaum
        public float sourceEndY;
        public float targetEndX; //Endpunkt xy des Beziers im Targetbaum
        public float targetEndY;
        public float currentEndX; //Endpunkt xy des Beziers im aktuellen Schritt
        public float currentEndY;
        public string wordFrom; //Das Wort, von dem der Bezier weg geht
        public string wordTo; //Das Wort, zu dem der Bezier fuehrt
        public float wordToSourceX; //Die xy Koordinaten des wordTo-Worts im Sourcebaum (fuer eindeutige Identifikation Wort+Koordinaten noetig)
        public float wordToSourceY;

        /**
         * Constructs this node.
         */
        public DisplayNodeBezier(float b0X, float b0Y, float b3X, float b3Y, string _wordFrom, string _wordTo, float _wordToSourceX, float _wordToSourceY)
        {
            sourceBeginX = b0X;
            sourceBeginY = b0Y;
            sourceEndX = b3X;
            sourceEndY = b3Y;
            wordTo = _wordTo;
            wordFrom = _wordFrom;
            wordToSourceX = _wordToSourceX;
            wordToSourceY = _wordToSourceY;
        }
    }

    /**
     * A list for animating displayed nodes.
     */
    public class DisplayList
    {
        protected LinkedList<DisplayNodeText> listWords; //Liste aller Wort-Nodes
        protected LinkedList<DisplayNodeBezier> listBeziers; //Liste aller Bezier-Nodes

        /**
         * Constructor - allocates an empty list.
         */
        public DisplayList()
        {
            listWords = new LinkedList<DisplayNodeText>();
            listBeziers = new LinkedList<DisplayNodeBezier>();
        }

        /**
         * Renders the DisplayList.
         * @param g The graphic object for drawing and measuring strings.
         * @param panelWidth Width of the drawing panel.
         * @param panelHeight Height of the drawing panel.
         * @param zoomFactor Zoom factor for the zoom window.
         * @param zoomWindowLeft Zoom window left coordinate.
         * @param zoomWindowTop Zoom window top coordinate.
         */
        public void render(Graphics g, int panelWidth, int panelHeight, float zoomFactor, float zoomWindowLeft, float zoomWindowTop)
        {
            //Errechne zooming parameter
            float zoomWindowHalfWidth = panelWidth / 2.0f;
            float zoomWindowHalfHeight = panelHeight / 2.0f;
            zoomWindowLeft += zoomWindowHalfWidth;
            zoomWindowTop += zoomWindowHalfHeight;

            //Rendere alle Text-Knoten an ihrer aktuellen Position
            foreach (DisplayNodeText t in listWords)
            {
                Font currentFontZoom = new Font(WordTree.font.FontFamily, t.currentFontsize * zoomFactor);
                Point childPositionZoom = new Point((int)((t.currentBeginX - zoomWindowLeft) * zoomFactor + zoomWindowHalfWidth), (int)((t.currentBeginY - zoomWindowTop) * zoomFactor + zoomWindowHalfHeight));
                if(currentFontZoom.Size>0.5)
                    g.DrawString(t.word, currentFontZoom, new SolidBrush(Color.Black), childPositionZoom);

            }

            //Rendere alle Bezierknoten an ihrer aktuellen Position
            foreach (DisplayNodeBezier b in listBeziers)
            {
                //now draw line:
                PointF b0 = new PointF(b.currentBeginX, b.currentBeginY);
                Point b0Zoom = new Point((int)((b0.X - zoomWindowLeft) * zoomFactor + zoomWindowHalfWidth), (int)((b0.Y - zoomWindowTop) * zoomFactor + zoomWindowHalfHeight));
                PointF b3 = new PointF(b.currentEndX, b.currentEndY);
                Point b3Zoom = new Point((int)((b3.X - zoomWindowLeft) * zoomFactor + zoomWindowHalfWidth), (int)((b3.Y - zoomWindowTop) * zoomFactor + zoomWindowHalfHeight));

                PointF b1 = new PointF((b0.X + b3.X) / 2, b0.Y);
                Point b1Zoom = new Point((int)((b1.X - zoomWindowLeft) * zoomFactor + zoomWindowHalfWidth), (int)((b1.Y - zoomWindowTop) * zoomFactor + zoomWindowHalfHeight));
                PointF b2 = new PointF(((b0.X + b3.X) / 2), b3.Y);
                Point b2Zoom = new Point((int)((b2.X - zoomWindowLeft) * zoomFactor + zoomWindowHalfWidth), (int)((b2.Y - zoomWindowTop) * zoomFactor + zoomWindowHalfHeight));

                g.DrawBezier(new Pen(new SolidBrush(Color.FromArgb(80, Color.Red))), b0Zoom, b1Zoom, b2Zoom, b3Zoom);
            }
        }

        /**
         * Updates the animation according to the input parameters.
         * @param param 0-1 = percent of the animation
         */
        public void update(float param)
        {
            foreach (DisplayNodeText t in listWords)
            {
                t.currentFontsize = param * t.targetFontsize + (1.0f - param) * t.sourceFontsize;
                t.currentBeginX =   param * t.targetBeginX +   (1.0f - param) * t.sourceBeginX;
                t.currentBeginY =   param * t.targetBeginY +   (1.0f - param) * t.sourceBeginY;
            }

            foreach (DisplayNodeBezier b in listBeziers)
            {
                b.currentBeginX = param * b.targetBeginX + (1.0f - param) * b.sourceBeginX;
                b.currentBeginY = param * b.targetBeginY + (1.0f - param) * b.sourceBeginY;
                b.currentEndX =   param * b.targetEndX +   (1.0f - param) * b.sourceEndX;
                b.currentEndY =   param * b.targetEndY +   (1.0f - param) * b.sourceEndY;
            }
        }

        /**
         * Merges two DisplayLists. The current DisplayLists only contains the positions in th source tree.
         * @param targetList List to merge.
         */
        public void mergeTargetList(DisplayList targetList)
        {
            //Suche fuer jedes Text-node in der Sourcelist ...
            foreach(DisplayNodeText src in listWords)
            {
                // ... das zugehoerige Text-node im Targetbaum (targetList)
                foreach (DisplayNodeText tar in targetList.listWords)
                {
                    //Falls es gefunden wurde, uebernimmm Position und Fontsize aus der Targetlist und speichere sie 
                    //auch bei dem source-Element ab. (somit wird fuer dieses Element die Interpolation dann moeglich)
                    if (src.word.Equals(tar.word))
                    {
                        src.targetBeginX = tar.sourceBeginX;
                        src.targetBeginY = tar.sourceBeginY;
                        src.targetFontsize = tar.sourceFontsize;
                        targetList.listWords.Remove(tar); //Wichtig: Entferne bereits verwendetes Element aus Targetlist
                        break;
                    }
                }
            }

            //Suche nun fuer jedes Bezier-Node der Sourcelist...
            foreach (DisplayNodeBezier src in listBeziers)
            {
                bool foundCorrespondingBezier = false;

                // .. das zuegehoerige Bezier in der targetList und speichere Targetkoordinaten beim Source-Element
                foreach (DisplayNodeBezier tar in targetList.listBeziers)
                {
                    if (src.wordFrom.Equals(tar.wordFrom) && src.wordTo.Equals(tar.wordTo))
                    {
                        src.targetBeginX = tar.sourceBeginX;
                        src.targetBeginY = tar.sourceBeginY;
                        
                        src.targetEndX = tar.sourceEndX;
                        src.targetEndY = tar.sourceEndY;

                        foundCorrespondingBezier = true;
                        targetList.listBeziers.Remove(tar); //Entferne das gefundene, verwendete Bezier
                        break;
                    }   
                }

                /*Sollte das Bezier allerdings im targetBaum nicht mehr vorhanden sein, such nach 
                * den woertern (und koordinaten)wo es sein sollte und interpoliere spaeter dorthin*/
                if (!foundCorrespondingBezier)
                {
                    foreach (DisplayNodeText wortNode in listWords)
                    {
                        if (src.wordTo.Equals(wortNode.word) && src.wordToSourceX == wortNode.sourceBeginX && src.wordToSourceY == wortNode.sourceBeginY)
                        {
                            src.targetBeginX = wortNode.targetBeginX;
                            src.targetBeginY = wortNode.targetBeginY + (wortNode.targetFontsize *0.7f);
                            src.targetEndX = wortNode.targetBeginX; //auf einen punkt zusammenziehen
                            src.targetEndY = wortNode.targetBeginY + (wortNode.targetFontsize *0.7f);
                         }
                    }
                } 
            }
            return;
        }

        /**
         * Reverses the animation.
         */
        public void reverseSourceTarget()
        {
            foreach (DisplayNodeText t in listWords)
            {
                float temp;
                temp = t.sourceBeginX;
                t.sourceBeginX = t.targetBeginX;
                t.targetBeginX = temp;

                temp = t.sourceBeginY;
                t.sourceBeginY = t.targetBeginY;
                t.targetBeginY = temp;

                temp = t.sourceFontsize;
                t.sourceFontsize = t.targetFontsize;
                t.targetFontsize = temp;
            }

            foreach (DisplayNodeBezier b in listBeziers)
            {
                float temp;
                temp = b.sourceBeginX;
                b.sourceBeginX = b.targetBeginX;
                b.targetBeginX = temp;

                temp = b.sourceBeginY;
                b.sourceBeginY = b.targetBeginY;
                b.targetBeginY = temp;

                temp = b.sourceEndX;
                b.sourceEndX = b.targetEndX;
                b.targetEndX = temp;

                temp = b.sourceEndY;
                b.sourceEndY = b.targetEndY;
                b.targetEndY = temp;
            }
        }

        /**
         * Add a source text element.
         */
        public void addSourceTextElement(DisplayNodeText elem)
        {
            listWords.AddLast(elem);
        }

        /**
         * Add a source bezier element.
         */
        public void addSourceBezierElement(DisplayNodeBezier elem)
        {
            listBeziers.AddLast(elem);
        }

    }

}