﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing;

namespace WordTree
{
    /**
     * One Node of the WordTree.
     */
    public class Node
    {
        private String word;
        public LinkedList<Node> children; //occurence sorted
        private LinkedList<Node> alphabeticallySortedChildren;
        private LinkedList<Node> frequencySortedChildren;
        private LinkedList<Node> currentChildren;
        private bool markYellow = false;
        private PointF lastPosition;

        private bool drawSolid;
        public bool doNotInsertIntoDisplayList = false;

        public int leafCount;

        /**
         * Copies the word and the markYellow property to a new Node.
         */
        public Node clone()
        {
            Node n = new Node(word);
            n.markYellow = markYellow;
            return n;
        }

        /**
         * Default constructor
         */
        public Node(String _word)
        {
            word = _word;
            leafCount = 0;
            children = new LinkedList<Node>();
            alphabeticallySortedChildren = new LinkedList<Node>();
            frequencySortedChildren = new LinkedList<Node>();
            drawSolid = false;
            switch (WordTree.sortMode)
            {
                case 0:
                    currentChildren = children;
                    break;
                case 1:
                    currentChildren = alphabeticallySortedChildren;
                    break;
                case 2:
                    currentChildren = frequencySortedChildren;
                    break;
            }
        }

        /**
         * Get the position where this Node was drawn in the last render call
         */
        public PointF getLastPosition()
        {
            return lastPosition;
        }

        /**
         * Set the last position.
         */
        public void setLastPosition(PointF lastPosition)
        {
            this.lastPosition = lastPosition;
        }

        /**
         * Mark this node yellow.
         */
        public void setYellow(bool value)
        {
            markYellow = value;
        }

        /**
         * Add a child Node to this Node (simple string).
         */
        public void addChild(String word)
        {
            addChild(new Node(word));
        }

        /**
         * Add a child Node to this Node.
         */
        public void addChild(Node node)
        {
            children.AddLast(node);

            LinkedListNode<Node> n = alphabeticallySortedChildren.First;
            while (n != null)
            {
                if (n.Value.getWord().CompareTo(node.getWord()) > 0)
                {
                    alphabeticallySortedChildren.AddBefore(n, children.Last.Value);
                    break;
                }
                n = n.Next;
            }

            if (n == null)
                alphabeticallySortedChildren.AddLast(children.Last.Value);

        }

        /**
         * Gets the word represented by this Node.
         */
        public String getWord()
        {
            return word;
        }

        /**
         * Add a sentence to this node and automatically generate all subnodes.
         */
        public bool addSentence(String[] sentence)
        {
            if (sentence.Length == 0)
                return false;

            foreach (Node n in children)
            {
                if (CaseInSensitiveEqualityComparer.EqualsStatic(n.getWord(), sentence[0]))
                {
                    if (n.addSentence(sentence.Skip(1).ToArray()))
                    {
                        n.leafCount++;
                        return true;
                    }
                    else
                        return false;

                }
            }

            bool increaseLeaves = (children.Count > 0);

            Node current = new Node(sentence[0]);
            this.addChild(current);

            for (int i = 1; i < sentence.Length; i++)
            {
                Node insertNode = new Node(sentence[i]);
                current.addChild(insertNode);
                current.leafCount++;
                current = insertNode;
            }

            return increaseLeaves;
        }

        /**
         * Draws all Children of this Node.
         * @param g The graphic object for drawing and measuring strings.
         * @param wordTree The parent WordTree.
         * @param rootPosition The position of the root element.
         * @param xEndOfPrevious x coordinate of this node's parent node.
         * @param yMin Minimum y of the target drawing region.
         * @param yMax Maximum y of the target drawing region.
         * @param parentFontSize Font size of the parent node.
         * @param zoomFactor Zoom factor for the zoom window.
         * @param zoomWindowLeft Zoom window left coordinate.
         * @param zoomWindowTop Zoom window top coordinate.
         * @param zoomWindowHalfWidth Zoom window half width.
         * @param zoomWindowHalfHeight Zoom window half height.
         * @param solid Draw everything solid?
         * @param displayList Render to a DisplayList.
         * @param visible Draw anything?
         * @param drawMarkedChildrenAlpha Draw the previously marked children with the class' alpha value.
         */
        public void drawAllChildren(Graphics g, WordTree wordTree, PointF rootPosition, float xEndOfPrevious, float yMin, float yMax, float parentFontSize, float zoomFactor, float zoomWindowLeft, float zoomWindowTop, float zoomWindowHalfWidth, float zoomWindowHalfHeight, bool solid, DisplayList displayList, bool visible, bool drawMarkedChildrenAlpha)
        {
            //wurde bereits ein hoeher gelegenes ge-clipped? dann brauchst nix mehr zeichnen.
            if (wordTree.clippedBefore && displayList == null)
                return;

            //leaf level?
            if (currentChildren.Count == 0)
                return;

            bool clippedBefore = false;

            //an welchem y-wert hat die vorige zelle geendet
            float yEndPrev = yMin;

            float maxLeaves = 1;

            foreach (Node child in currentChildren)
            {
                if (child.children.Count > maxLeaves)
                    maxLeaves = child.leafCount;
            }


            IEnumerator<Node> enumer = currentChildren.GetEnumerator();
            //schleife ueber alle kinder:
            while(enumer.MoveNext())
            {
                //jener vertikale platz in pixel, der den kindern des aktuellen nodes zufaellt
                float verticalSpace = ((yMax - yMin) * Math.Max(1, enumer.Current.leafCount) / Math.Max(1, leafCount));

                //calculate vertical start and endof this child's cell:
                float yStart = yEndPrev;// yMin + i * verticalSpace;
                yEndPrev = yStart + verticalSpace;

                PointF childPosition;
                Font currentFont, currentFontZoom;

                //draw child word without line, if only child:
                if (currentChildren.Count == 1)
                {
                    currentFont = new Font(WordTree.font.FontFamily, parentFontSize);
                    currentFontZoom = new Font(WordTree.font.FontFamily, currentFont.SizeInPoints * zoomFactor);
                    childPosition = new PointF((xEndOfPrevious), (yStart + verticalSpace / 2) - currentFont.GetHeight(g) / 2);
                    Point childPositionZoom = new Point((int)((childPosition.X - zoomWindowLeft) * zoomFactor + zoomWindowHalfWidth), (int)((childPosition.Y - zoomWindowTop) * zoomFactor + zoomWindowHalfHeight));
                    
                    enumer.Current.setLastPosition(childPosition);

                    //clipping:
                    if (displayList != null || childPositionZoom.Y < WordTree.yMax) //ist es oberhalb der unteren kante?
                    {
                        if (displayList != null || childPositionZoom.Y >= -1 * currentFontZoom.GetHeight()) //ist es unterhalb der oberen kante?
                        {

                            Brush brush;
                            int alpha = 255;

                            if (!enumer.Current.Equals(wordTree.lastSplitNode) && !solid && !enumer.Current.drawSolid)
                          	  alpha = wordTree.animationAlpha;
                            
                            if(drawMarkedChildrenAlpha && enumer.Current.doNotInsertIntoDisplayList)
                                alpha = wordTree.animationAlpha;

						    if (drawMarkedChildrenAlpha && !enumer.Current.doNotInsertIntoDisplayList)
                                alpha = 255;

  							if (displayList != null && alpha == 255 && !enumer.Current.doNotInsertIntoDisplayList)
                                displayList.addSourceTextElement(new DisplayNodeText(enumer.Current.getWord(), childPosition, currentFont.Size));

                            if (enumer.Current.leafCount <= 1)
                                brush = new SolidBrush(Color.FromArgb(alpha, Color.Gray));
                            else
                                brush = new SolidBrush(Color.FromArgb(alpha, Color.Black));

                            if (visible && currentFontZoom.GetHeight(g) > 0.5)
                            {
  								 if (enumer.Current.markYellow)
                                {
                                    float lengthOfWordForMark = g.MeasureString(enumer.Current.getWord(), currentFontZoom).Width;
                                    g.FillRectangle(Brushes.Yellow, childPositionZoom.X,
                                                                            childPositionZoom.Y,
                                                                            lengthOfWordForMark,
                                                                            currentFontZoom.Height);
                                }
                                if (WordTree.showLeafCount)
                                    g.DrawString(enumer.Current.getWord() + enumer.Current.leafCount, currentFontZoom, brush, childPositionZoom);
                                else
                                    g.DrawString(enumer.Current.getWord(), currentFontZoom, brush, childPositionZoom);
                            }
                            wordTree.wordsRendered++;
                        }

                    }
                }
                else
                {
                    if (WordTree.fontSizeCoding == 1)
                    {
                        currentFont = new Font(WordTree.font.FontFamily, parentFontSize * WordTree.fontScale);
                        currentFont = new Font(currentFont.FontFamily, Math.Min(parentFontSize, currentFont.Size + WordTree.resizeableFontSize * (float)Math.Max(1, enumer.Current.leafCount) / (float)maxLeaves));
                    }
                    else
                    {
                        currentFont = new Font(WordTree.font.FontFamily, (float)Math.Max(1.0, Math.Sqrt((double)((double)(WordTree.font.Size * enumer.Current.leafCount)) / (double)wordTree.root.leafCount)));
                    }

                    childPosition = new PointF(xEndOfPrevious + (float)(WordTree.bezierScaling*wordTree.numLeaves*WordTree.hspace * Math.Sqrt((float)leafCount / (float)wordTree.numLeaves)), (yStart + verticalSpace / 2.0f) - currentFont.GetHeight(g) / 2);
                    enumer.Current.setLastPosition(childPosition);
                    currentFontZoom = new Font(WordTree.font.FontFamily, currentFont.SizeInPoints * zoomFactor);

                    Point childPositionZoom = new Point((int)((childPosition.X - zoomWindowLeft) * zoomFactor + zoomWindowHalfWidth), (int)((childPosition.Y - zoomWindowTop) * zoomFactor + zoomWindowHalfHeight));

                    //clipping:
                    if (displayList != null || childPositionZoom.Y < WordTree.yMax) //ist es oberhalb der unteren kante?
                    {
                        if (displayList != null || childPositionZoom.Y >= -1 * currentFontZoom.GetHeight()) //ist es unterhalb der oberen kante?
                        {
                            Brush brush;

                            if (solid || enumer.Current.Equals(wordTree.lastSplitNode) || enumer.Current.drawSolid)
                            {
                                brush = new SolidBrush(Color.Black);
                                if (displayList != null && !enumer.Current.doNotInsertIntoDisplayList)
                                    displayList.addSourceTextElement(new DisplayNodeText(enumer.Current.getWord(), childPosition, currentFont.Size));
                            }
                            else
                            {
                                if(!drawMarkedChildrenAlpha)
                                    brush = new SolidBrush(Color.FromArgb(wordTree.animationAlpha, Color.Black));
                                else
                                    if(enumer.Current.doNotInsertIntoDisplayList)
                                        brush = new SolidBrush(Color.FromArgb(wordTree.animationAlpha, Color.Black));
                                    else
                                        brush = new SolidBrush(Color.FromArgb(255, Color.Black));

                                if (displayList != null && wordTree.animationAlpha == 255 && !enumer.Current.doNotInsertIntoDisplayList)
                                    displayList.addSourceTextElement(new DisplayNodeText(enumer.Current.getWord(), childPosition, currentFont.Size));
                            }

                            if (visible && currentFontZoom.GetHeight(g) > 0.5)
                            {
  								if (enumer.Current.markYellow)
                                {
                                    float lengthOfWordForMark = g.MeasureString(enumer.Current.getWord(), currentFontZoom).Width;
                                    g.FillRectangle(Brushes.Yellow, childPositionZoom.X,
                                                                            childPositionZoom.Y,
                                                                            lengthOfWordForMark,
                                                                            currentFontZoom.Height);


                                }

                                if (WordTree.showLeafCount)
                                    g.DrawString(enumer.Current.getWord() + enumer.Current.leafCount, currentFontZoom, brush, childPositionZoom);
                                else
                                    g.DrawString(enumer.Current.getWord(), currentFontZoom, brush, childPositionZoom);
                            }
                            wordTree.wordsRendered++;
                        }
                    }
                    else
                    {
                        clippedBefore = true;
                    }

                    //now draw line:
                    PointF b0 = new PointF((xEndOfPrevious), rootPosition.Y);
                    Point b0Zoom = new Point((int)((b0.X - zoomWindowLeft) * zoomFactor + zoomWindowHalfWidth), (int)((b0.Y - zoomWindowTop) * zoomFactor + zoomWindowHalfHeight));
                    PointF b3 = new PointF(childPosition.X - 0, childPosition.Y + currentFont.GetHeight(g) / 2);
                    Point b3Zoom = new Point((int)((b3.X - zoomWindowLeft) * zoomFactor + zoomWindowHalfWidth), (int)((b3.Y - zoomWindowTop) * zoomFactor + zoomWindowHalfHeight));

                    PointF b1 = new PointF(((childPosition.X - 0 + xEndOfPrevious) / 2), rootPosition.Y);
                    Point b1Zoom = new Point((int)((b1.X - zoomWindowLeft) * zoomFactor + zoomWindowHalfWidth), (int)((b1.Y - zoomWindowTop) * zoomFactor + zoomWindowHalfHeight));
                    PointF b2 = new PointF(((childPosition.X - 0 + xEndOfPrevious) / 2), childPosition.Y + currentFont.GetHeight(g) / 2);
                    Point b2Zoom = new Point((int)((b2.X - zoomWindowLeft) * zoomFactor + zoomWindowHalfWidth), (int)((b2.Y - zoomWindowTop) * zoomFactor + zoomWindowHalfHeight));

                    float bezTop = Math.Min(b0Zoom.Y, b3Zoom.Y);
                    float bezBottom = Math.Max(b0Zoom.Y, b3Zoom.Y);

                    int alpha = 80;

                    if (!solid && !enumer.Current.drawSolid)
                        alpha = (int)((80.0f / 255.0f) * (float)wordTree.animationAlpha);
                    
                    if (drawMarkedChildrenAlpha && enumer.Current.doNotInsertIntoDisplayList)
                        alpha = (int)((80.0f / 255.0f) * (float)wordTree.animationAlpha);

                    if (drawMarkedChildrenAlpha && !enumer.Current.doNotInsertIntoDisplayList)
                        alpha = 80;

                    if (displayList != null && alpha == 80 && !enumer.Current.doNotInsertIntoDisplayList)
                        displayList.addSourceBezierElement(new DisplayNodeBezier(b0.X, b0.Y, b3.X, b3.Y, this.getWord(), enumer.Current.getWord(), childPosition.X, childPosition.Y));

                    if (visible && (bezTop >= -1 || bezBottom <= WordTree.yMax))
                        g.DrawBezier(new Pen(new SolidBrush(Color.FromArgb(alpha, Color.Red))), b0Zoom, b1Zoom, b2Zoom, b3Zoom);

                }

                float lengthOfWord = g.MeasureString(enumer.Current.getWord() + " ", currentFont).Width;
                //debug vertical space output:
                //g.DrawRectangle(new Pen(new SolidBrush(Color.Black)), new Rectangle((int)childPosition.X, (int)yStart, (int)(lengthOfWord), (int)verticalSpace));

                if (enumer.Current.children.Count != 1)
                    enumer.Current.drawAllChildren(g, wordTree, new PointF(childPosition.X, childPosition.Y + (currentFont.GetHeight(g) / 2)), childPosition.X + lengthOfWord, yStart, yStart + verticalSpace, parentFontSize * WordTree.fontScale, zoomFactor, zoomWindowLeft, zoomWindowTop, zoomWindowHalfWidth, zoomWindowHalfHeight, solid || enumer.Current.Equals(wordTree.lastSplitNode), displayList, visible, drawMarkedChildrenAlpha);
                else
                    enumer.Current.drawAllChildren(g, wordTree, new PointF(childPosition.X, childPosition.Y + (currentFont.GetHeight(g) / 2)), childPosition.X + lengthOfWord, yStart, yStart + verticalSpace, currentFont.Size, zoomFactor, zoomWindowLeft, zoomWindowTop, zoomWindowHalfWidth, zoomWindowHalfHeight, solid || enumer.Current.Equals(wordTree.lastSplitNode), displayList, visible, drawMarkedChildrenAlpha);

                if (clippedBefore)
                    wordTree.clippedBefore = true;
            }
        }

        /**
         * Finds a node.
         * @param g The graphic object for drawing and measuring strings.
         * @param wordTree The parent WordTree.
         * @param x 
         * @param y
         * @param rootPosition The position of the root element.
         * @param xEndOfPrevious x coordinate of this node's parent node.
         * @param yMin Minimum y of the target drawing region.
         * @param yMax Maximum y of the target drawing region.
         * @param parentFontSize Font size of the parent node.
         * @param zoomFactor Zoom factor for the zoom window.
         * @param zoomWindowLeft Zoom window left coordinate.
         * @param zoomWindowTop Zoom window top coordinate.
         * @param zoomWindowHalfWidth Zoom window half width.
         * @param zoomWindowHalfHeight Zoom window half height.
         */
        public LinkedList<String> findNode(Graphics g, WordTree wordTree, int x, int y, PointF rootPosition, float xEndOfPrevious, float yMin, float yMax, float parentFontSize, float zoomFactor, float zoomWindowLeft, float zoomWindowTop, float zoomWindowHalfWidth, float zoomWindowHalfHeight)
        {
            //leaf level?
            if (children.Count == 0)
                return null;

            //an welchem y-wert hat die vorige zelle geendet
            float yEndPrev = yMin;

            float maxLeaves = 1;

            foreach (Node child in currentChildren)
            {
                if (child.children.Count > maxLeaves)
                    maxLeaves = child.leafCount;
            }

            //schleife ueber alle kinder:
            for (int i = 0; i < children.Count; i++)
            {
                //jener vertikale platz in pixel, der den kindern des aktuellen nodes zufaellt
                //rem: replaced children by currentChildren
                float verticalSpace = ((yMax - yMin) * Math.Max(1, currentChildren.ElementAt(i).leafCount) / leafCount);

                //calculate vertical start and endof this child's cell:
                float yStart = yEndPrev;// yMin + i * verticalSpace;
                yEndPrev = yStart + verticalSpace;

                PointF childPosition;
                Font currentFont = WordTree.font, currentFontZoom;

                //currentFont = new Font(currentFont.FontFamily, currentFont.Size * (float)Math.Max(1, currentChildren.ElementAt(i).leafCount) / (float)WordTree.numLeaves);
                //currentFontZoom = new Font(WordTree.font.FontFamily, currentFont.SizeInPoints * zoomFactor);

                if (children.Count == 1)
                {
                    currentFont = new Font(WordTree.font.FontFamily, parentFontSize);
                    currentFontZoom = new Font(WordTree.font.FontFamily, currentFont.SizeInPoints * zoomFactor);
                    childPosition = new PointF((xEndOfPrevious), (yStart + verticalSpace / 2) - currentFont.GetHeight(g) / 2);

                }
                else
                {
                    currentFont = new Font(WordTree.font.FontFamily, parentFontSize * WordTree.fontScale);
                    currentFont = new Font(currentFont.FontFamily, Math.Min(parentFontSize, currentFont.Size + WordTree.resizeableFontSize * (float)Math.Max(1, currentChildren.ElementAt(i).leafCount) / (float)maxLeaves));

                    currentFontZoom = new Font(WordTree.font.FontFamily, currentFont.SizeInPoints * zoomFactor);
                      childPosition = new PointF(xEndOfPrevious + (float)(WordTree.bezierScaling * wordTree.numLeaves * WordTree.hspace * Math.Sqrt((float)leafCount / (float)wordTree.numLeaves)), (yStart + verticalSpace / 2.0f) - currentFont.GetHeight(g) / 2);
                    //childPosition = new PointF(xEndOfPrevious + (float)(WordTree.hspace * Math.Sqrt((float)leafCount / (float)wordTree.numLeaves)), (yStart + verticalSpace / 2.0f) - currentFont.GetHeight(g) / 2);
                }

                Point childPositionZoom = new Point((int)((childPosition.X - zoomWindowLeft) * zoomFactor + zoomWindowHalfWidth), (int)((childPosition.Y - zoomWindowTop) * zoomFactor + zoomWindowHalfHeight));

                float lengthOfWord = g.MeasureString(currentChildren.ElementAt(i).getWord(), currentFont).Width;
                float lengthOfWordZoom = g.MeasureString(currentChildren.ElementAt(i).getWord(), currentFontZoom).Width;

                //clipping:
                if (childPositionZoom.Y < WordTree.yMax) //ist es oberhalb der unteren kante?
                {
                    if (childPositionZoom.Y >= -1 * currentFontZoom.GetHeight(g)) //ist es unterhalb der oberen kante?
                    {
                        //click test:
                        if (x >= childPositionZoom.X && x <= (childPositionZoom.X + lengthOfWordZoom) && y >= childPositionZoom.Y && y <= childPositionZoom.Y + currentFontZoom.GetHeight(g))
                        {
                            LinkedList<string> resultString = new LinkedList<string>();
                            //rem: replaced children by currentChildren
                            resultString.AddFirst(currentChildren.ElementAt(i).getWord());
                            resultString.AddFirst(this.word);

                            wordTree.lastSplitNode = currentChildren.ElementAt(i);
                            currentChildren.ElementAt(i).drawSolid = true; //clicked node
                            this.drawSolid = true;// is on path to clicked node

                            return resultString;
                        }
                    }
                }


                LinkedList<string> subtreeResult;
                if (currentChildren.ElementAt(i).children.Count != 1)
                    subtreeResult = currentChildren.ElementAt(i).findNode(g, wordTree, x, y, new PointF(childPosition.X, childPosition.Y + (currentFont.GetHeight(g) / 2)), childPosition.X + lengthOfWord, yStart, yStart + verticalSpace, parentFontSize * WordTree.fontScale, zoomFactor, zoomWindowLeft, zoomWindowTop, zoomWindowHalfWidth, zoomWindowHalfHeight);
                else
                    subtreeResult = currentChildren.ElementAt(i).findNode(g, wordTree, x, y, new PointF(childPosition.X, childPosition.Y + (currentFont.GetHeight(g) / 2)), childPosition.X + lengthOfWord, yStart, yStart + verticalSpace, currentFont.Size, zoomFactor, zoomWindowLeft, zoomWindowTop, zoomWindowHalfWidth, zoomWindowHalfHeight);

                if (subtreeResult != null)
                {
                    subtreeResult.AddFirst(this.word);
                    this.drawSolid = true; //is on path to clicked node
                    return subtreeResult;
                }
            }

            return null;
        }

        /**
         * Changes the sorting mode.
         * 0 - As elements occur in the document
         * 1 - Alphabetically
         * 2 - By frequency
         */
        public void resort(int mode)
        {
            switch (mode)
            {
                case 0:
                    currentChildren = children;
                    break;
                case 1:
                    currentChildren = alphabeticallySortedChildren;
                    break;
                case 2:
                    currentChildren = frequencySortedChildren;
                    break;
            }

            foreach (Node n in children)
                n.resort(mode);
        }

        /**
         * Creates the frequency sorting for this node.
         */
        public void createFrequencySorting()
        {
            foreach (Node n in children)
            {
                LinkedListNode<Node> comp = frequencySortedChildren.First;
                while (comp != null)
                {
                    if (comp.Value.leafCount < n.leafCount)
                    {
                        frequencySortedChildren.AddBefore(comp, n);
                        break;
                    }
                    comp = comp.Next;
                }

                if (comp == null)
                    frequencySortedChildren.AddLast(n);

                n.createFrequencySorting();
            }
        }
    }

    /**
     * The WordTree.
     */
    public class WordTree
    {
        public Node root;
        public int numLeaves;

        public static bool busy;

        public Node lastSplitNode = null;
        public int animationAlpha = 255;

        public int wordsRendered;
        public bool clippedBefore;
        public static bool showLeafCount = false;
        public static float resizeableFontSize = 14.0f;
        public static float fontScale = 0.45f;

        public Point rootPosition;

        public static int xMax; //breite des canvas-rechtecks
        public static int yMax; //hoehe des canvas-rechtecks

        public static Font font = new Font("Verdana", 24); //globaler font. Falls baum zu gross wird, wird schriftgroesse angepasst.
        public static Brush brush = new SolidBrush(Color.Black); //globaler brush.

        public static float hspace = 20; //horizontaler space in pixel zwischen zwei aufeinanderfolgenden worten, falls verzweigung

        public static float bezierScaling;

        public static int fontSizeCoding = 1;

        public static int sortMode; //0=occurence, 1=alphabet 2=frequency

        /**
         * Private constructor.
         */
        private WordTree()
        {
        }

        /**
         * Constructs a WordTree with an initial string e.g. "My name is rudolf"
         */
        public WordTree(String searchString)
        {
            String[] nodes = searchString.Split(new char[] { ' ' });

            root = new Node(nodes[0]);
           
            Node current = root;
            for (int i = 1; i < nodes.Length; i++)
            {
                Node insertNode = new Node(nodes[i]);
                current.addChild(insertNode);
                current.leafCount++;
                current = insertNode;
            }

            numLeaves = 1;
        }

       /**
        * Calcs the position of the root node.
        */
        public void calcRootPosition()
        {
            float yMax = (int)(root.leafCount * WordTree.font.Size * 2 * fontScale);
            rootPosition = new Point(0, (int)(yMax - WordTree.font.GetHeight()) / 2);
        }

        /**
         * Adds a sentence to the WordTree.
         */
        public void addSentence(String sentence)
        {
            addSentence(sentence.Split(new char[] { ' ' }));
        }

        /**
         * Adds a sentence to the WordTree (already separated in an array).
         */
        public void addSentence(String[] sentence)
        {
            if (CaseInSensitiveEqualityComparer.EqualsStatic(root.getWord(), sentence[0]))
            {
                if (root.addSentence(sentence.Skip(1).ToArray()))
                    numLeaves++;
            }

            root.leafCount = numLeaves;
        }

        /**
         * Draws all Children of this tree.
         * @param g The graphic object for drawing and measuring strings.
         * @param panelWidth Width of the drawing panel.
         * @param panelHeight Height of the drawing panel.
         * @param zoomFactor Zoom factor for the zoom window.
         * @param zoomWindowLeft Zoom window left coordinate.
         * @param zoomWindowTop Zoom window top coordinate.
         * @param zoomWindowHalfWidth Zoom window half width.
         * @param zoomWindowHalfHeight Zoom window half height.
         * @param displayList Render to a DisplayList.
         * @param visible Draw anything?
         * @param drawMarkedChildrenAlpha Draw the previously marked children with the class' alpha value.
         */
        public void render(Graphics g, int panelWidth, int panelHeight, float zoomFactor, float zoomWindowLeft, float zoomWindowTop, DisplayList displayList, bool visible, bool drawMarkedChildrenAlpha)
        {
            //canvas for tree (defined by rectangle, starting at origin)
            xMax = panelWidth;
            //yMax = (int)(root.leafCount * WordTree.font.Size * 2 * fontScale);
            float yMax = (int)(root.leafCount * WordTree.font.Size * 2 * fontScale);
            WordTree.yMax = panelHeight;

            float zoomWindowHalfWidth = panelWidth / 2.0f;
            float zoomWindowHalfHeight = panelHeight / 2.0f;

            zoomWindowLeft += zoomWindowHalfWidth;
            zoomWindowTop += zoomWindowHalfHeight;

            //draw canvas rectangle:
            if(visible)
                g.DrawRectangle(new Pen(brush), 0, 0, xMax, panelHeight);

            //draw root: 
            Font currentFont = new Font(WordTree.font.FontFamily, WordTree.font.SizeInPoints);
            Font zoomFont = new Font(currentFont.FontFamily, currentFont.SizeInPoints * zoomFactor);

            //Brush rootBrush = new SolidBrush(Color.FromArgb(animationAlpha, Color.Black));
            Brush rootBrush = new SolidBrush(Color.Black);

            //Point rootPosition;
            //rootPosition = new Point(0, (int)(yMax - WordTree.font.GetHeight()) / 2);

            float yOffset = rootPosition.Y - (int)(yMax - WordTree.font.GetHeight()) / 2;
   
            Point rootPositionZoom = new Point((int)((rootPosition.X - zoomWindowLeft) * zoomFactor + zoomWindowHalfWidth), (int)((rootPosition.Y - zoomWindowTop) * zoomFactor + zoomWindowHalfHeight));

            if (visible && zoomFont.GetHeight(g) > 0.5)
            {
                if (WordTree.showLeafCount)
                    g.DrawString(root.getWord() + numLeaves.ToString(), zoomFont, rootBrush, rootPositionZoom);
                else
                    g.DrawString(root.getWord(), zoomFont, rootBrush, rootPositionZoom);
            }

            if (displayList != null)
                displayList.addSourceTextElement(new DisplayNodeText(root.getWord(), rootPosition, currentFont.Size));

            wordsRendered = 0;
            clippedBefore = false;

            int lengthOfRootWord = TextRenderer.MeasureText(root.getWord(), currentFont).Width;
            
            root.drawAllChildren(g, this, new Point(rootPosition.X, rootPosition.Y + (int)(currentFont.GetHeight() / 2)), lengthOfRootWord, yOffset+0, yOffset+yMax, WordTree.font.Size, zoomFactor, zoomWindowLeft, zoomWindowTop, zoomWindowHalfWidth, zoomWindowHalfHeight, false, displayList, visible, drawMarkedChildrenAlpha);
           /* 
            try
            {
                if (visible)
                   g.DrawString(numLeaves.ToString() + " Vorkommnisse gefunden, " + wordsRendered.ToString() + " Wörter gerendert", new Font(currentFont.FontFamily, 10.0f), brush, new Point(0, 0));
            }
            catch (Exception)
            { }
			*/

        }

        /**
         * Gets the list of substrings in an area.
         * @param g The graphic object for drawing and measuring strings.
         * @param x Area x.
         * @param y Area y.
         * @param panelWidth Width of the drawing panel.
         * @param panelHeight Height of the drawing panel.
         * @param zoomFactor Zoom factor for the zoom window.
         * @param zoomWindowLeft Zoom window left coordinate.
         * @param zoomWindowTop Zoom window top coordinate.
         */
        public LinkedList<string> getSubtreeStringList(Graphics g, int x, int y, float panelWidth, float panelHeight, float zoomFactor, float zoomWindowLeft, float zoomWindowTop)
        {
            float xMax = panelWidth;
            float yMax = (int)(root.leafCount * WordTree.font.Size * 2 * fontScale);
            WordTree.yMax = (int)panelHeight;

            float zoomWindowHalfWidth = panelWidth / 2.0f;
            float zoomWindowHalfHeight = panelHeight / 2.0f;

            zoomWindowLeft += zoomWindowHalfWidth;
            zoomWindowTop += zoomWindowHalfHeight;

            Font currentFont = new Font(WordTree.font.FontFamily, WordTree.font.SizeInPoints);
            Font zoomFont = new Font(currentFont.FontFamily, currentFont.SizeInPoints * zoomFactor);

            //Point rootPosition = new Point(0, (int)(yMax - WordTree.font.GetHeight()) / 2);

            float yOffset = rootPosition.Y - (int)(yMax - WordTree.font.GetHeight()) / 2;

            Point rootPositionZoom = new Point((int)((rootPosition.X - zoomWindowLeft) * zoomFactor + zoomWindowHalfWidth), (int)((rootPosition.Y - zoomWindowTop) * zoomFactor + zoomWindowHalfHeight));
            int lengthOfRootWord = TextRenderer.MeasureText(root.getWord(), currentFont).Width;
            int lengthOfRootWordZoom = TextRenderer.MeasureText(root.getWord(), zoomFont).Width;

            if (x >= rootPositionZoom.X && x <= (rootPositionZoom.X + lengthOfRootWordZoom) && y >= rootPositionZoom.Y && y <= rootPositionZoom.Y + zoomFont.GetHeight(g))
            {
                LinkedList<string> resultString = new LinkedList<string>();
                resultString.AddFirst(root.getWord());
                lastSplitNode = root;
                return resultString;
            }

            return root.findNode(g, this, x, y, new Point(rootPosition.X, rootPosition.Y + (int)(WordTree.font.GetHeight() / 2)), lengthOfRootWord, yOffset + 0, yOffset +yMax, WordTree.font.Size, zoomFactor, zoomWindowLeft, zoomWindowTop, zoomWindowHalfWidth, zoomWindowHalfHeight);

        }

        /**
         * Changes the sorting mode.
         * 0 - As elements occur in the document
         * 1 - Alphabetically
         * 2 - By frequency
         */
        public void resort(int mode)
        {
            WordTree.sortMode = mode;
            root.resort(mode);
        }

        /**
         * Creates the frequency sorting for all nodes.
         */
        public void createFrequencySorting()
        {
            root.createFrequencySorting();
        }

        /**
         * Searches for a word.
         */
        public LinkedList<PointF> searchWord(string word)
        {
            LinkedList<PointF> positions = new LinkedList<PointF>();
            searchWordPrivate(word, root, positions);
            return positions;
        }

        /**
         * Searches for a word (private recursion).
         */
        private void searchWordPrivate(string word, Node n, LinkedList<PointF> positions)
        {
            foreach (Node subnode in n.children)
            {
                if (CaseInSensitiveEqualityComparer.EqualsStatic(subnode.getWord(), word))
                {
                    subnode.setYellow(true);
                    positions.AddLast(subnode.getLastPosition());
                }
                else
                    subnode.setYellow(false);
                searchWordPrivate(word, subnode, positions);
            }
        }

        /**
         * Removes unimportanted nodes.
         * @param percent 0-100
         */
        public WordTree removeUnimportantNodes(int percent)
        {
            WordTree newTree = new WordTree();

            newTree.root = root.clone();
            newTree.numLeaves = removeUnimportantNodesPrivate(percent, root, newTree.root);

            return newTree;
        }

        /**
         * Removes unimportanted nodes (private recursion).
         */
        private int removeUnimportantNodesPrivate(int percent, Node oldNode, Node newNode)
        {
            IEnumerator<Node> enumer = oldNode.children.GetEnumerator();
            int numAdded = 0;

            //schleife ueber alle kinder:
            while (enumer.MoveNext())
            {
                if ((enumer.Current.leafCount * 100) / oldNode.leafCount >= percent || oldNode.children.Count < 9)
                { 
                    Node newChildNode = enumer.Current.clone();
                    newNode.addChild(newChildNode);
                    numAdded += removeUnimportantNodesPrivate(percent, enumer.Current, newChildNode);
                }
            }

            // Wenigstens eins hinzufuegen, wenn Kinder da sind
            if (numAdded == 0 && oldNode.children.Count > 0)
            {
                Node newChildNode = enumer.Current.clone();
                newNode.addChild(newChildNode);
                numAdded += removeUnimportantNodesPrivate(percent, enumer.Current, newChildNode);
            }

            newNode.leafCount = numAdded;
            if (oldNode.children.Count == 0)
                return 1;
            else
                return numAdded;

        }

        /**
         * Marks all side nodes. I.e. all nodes over and under the words in array (graphically).
         */
        public void markAllSideNodes(string[] array)
        {
            markAllSideNodesPrivate(this.root, array, 1);
        }

        /**
         * Marks all side nodes. I.e. all nodes over and under the words in array (graphically, private recursion).
         */
        private void markAllSideNodesPrivate(Node n, string[] array, int index)
        {
            if (index >= array.Length)
                return;

            foreach (Node node in n.children)
            {
                if (node.getWord().Equals(array[index]))
                {
                    markSubtree(node, false);
                    markAllSideNodesPrivate(node, array, index + 1);
                }
                else
                    markSubtree(node, true);
            }
        }

        /**
         * Private mark a subtree with value)
         */
        private void markSubtree(Node n, bool value)
        {
            n.doNotInsertIntoDisplayList = value;
            foreach (Node node in n.children)
                markSubtree(node, value);
        }

        /**
         * Resets all markings.
         */
        public void resetSideNodeMarking()
        {
            markSubtree(root, false);
        }
    }



}

